import math
from abc import ABC, abstractmethod
from collections import defaultdict

from tarsier.ocr.types import ImageAnnotatorResponse


class OCRService(ABC):
    @abstractmethod
    def annotate(self, image_file: bytes) -> ImageAnnotatorResponse:
        pass

    @staticmethod
    def format_text(ocr_text: ImageAnnotatorResponse) -> str:
        # Initialize dimensions
        canvas_width = 200
        canvas_height = 100

        # Cluster tokens by line
        line_cluster = defaultdict(list)
        for annotation in ocr_text["words"]:
            # y = math.floor(annotation.midpoint_normalized[1] * canvas_height)
            y = round(annotation["midpoint_normalized"][1], 3)
            line_cluster[y].append(annotation)
        canvas_height = max(canvas_height, len(line_cluster))

        # find max line length
        max_line_length = max(
            sum([len(token["text"]) + 1 for token in line])
            for line in line_cluster.values()
        )
        canvas_width = max(canvas_width, max_line_length)

        # Create an empty canvas (list of lists filled with spaces)
        canvas = [[" " for _ in range(canvas_width)] for _ in range(canvas_height)]

        # Place the annotations on the canvas
        for i, (y, line_annotations) in enumerate(line_cluster.items()):
            # Sort annotations in this line by x coordinate
            line_annotations.sort(key=lambda e: e["midpoint_normalized"][0])

            last_x = 0  # Keep track of the last position where text was inserted
            for annotation in line_annotations:
                text = annotation["text"]

                x = math.floor(annotation["midpoint_normalized"][0] * canvas_width)

                # Move forward if there's an overlap
                x = max(x, last_x)

                # Check if the text fits; if not, move to next line (this is simplistic)
                if x + len(text) >= canvas_width:
                    continue  # TODO: extend the canvas_width in this case

                # Place the text on the canvas
                for j, char in enumerate(text):
                    canvas[i][x + j] = char

                # Update the last inserted position
                last_x = x + len(text) + 1  # +1 for a space between words

        # Delete all whitespace characters after the last non-whitespace character in each row
        canvas = [list("".join(row).rstrip()) for row in canvas]

        # Convert the canvas to a plaintext string
        page_text = "\n".join("".join(row) for row in canvas)
        page_text = page_text.strip()
        page_text = "-" * canvas_width + "\n" + page_text + "\n" + "-" * canvas_width
        page_text = page_text.replace("        ", "\t")

        return page_text



from typing import Any, Dict

from google.cloud import vision

from tarsier.ocr._base import OCRService
from tarsier.ocr.types import ImageAnnotatorResponse


class GoogleVisionOCRService(OCRService):
    def __init__(self, credentials: Dict[str, Any]):
        try:
            self.client = vision.ImageAnnotatorClient.from_service_account_info(
                credentials
            )
        except Exception:  # TODO: specify exception
            raise ValueError(
                "OCR client creation from credentials failed.\n"
                "Google your google cloud vision credentials can be created here:\n"
                "https://console.cloud.google.com/apis/api/vision.googleapis.com"
            )

        text_detection = vision.Feature()
        text_detection.type_ = vision.Feature.Type.TEXT_DETECTION  # type: ignore
        self._features = [text_detection]

    def annotate(self, image_file: bytes) -> ImageAnnotatorResponse:
        image = vision.Image()
        image.content = image_file

        request = vision.AnnotateImageRequest()
        request.image = image
        request.features = self._features

        res = self.client.annotate_image(request)  # TODO: make this async?

        annotations = res.text_annotations
        whole_text_box_max = annotations[0].bounding_poly.vertices[2]
        max_width = whole_text_box_max.x
        max_height = whole_text_box_max.y

        annotations_normed = []
        for text in annotations[1:]:
            box = text.bounding_poly.vertices

            # midpoint: average position betwen
            # the upper left location and lower
            # right position
            midpoint = ((box[2].x + box[0].x) / 2, (box[2].y + box[0].y) / 2)

            annotations_normed.append(
                {
                    "text": text.description,
                    "midpoint": midpoint,
                    "midpoint_normalized": (
                        midpoint[0] / max_width,
                        midpoint[1] / max_height,
                    ),
                }
            )

        annotations_normed = list(
            sorted(
                annotations_normed,
                key=lambda x: (
                    x["midpoint_normalized"][1],
                    x["midpoint_normalized"][0],
                ),
            )
        )

        return {"words": annotations_normed}  # type: ignore



from typing import Any

from playwright.async_api import Page as PageAsync
from playwright.sync_api import Page as PageSync

from tarsier.adapter._base import BrowserAdapter
from tarsier.adapter.types import ViewPortSize


class PlaywrightSyncAdapter(BrowserAdapter):
    STRIP_RETURN = "return window."

    def __init__(self, page: PageSync):
        self._page = page
        raise NotImplementedError(
            "Sync playwright is not yet supported. Please use the PlaywrightAsyncDriver instead."
        )

    async def run_js(self, js: str) -> Any:
        if js.startswith(self.STRIP_RETURN):
            js = js[len(self.STRIP_RETURN) :]

        return self._page.evaluate(js)

    async def take_screenshot(self) -> bytes:
        return self._page.screenshot(type="png")

    async def set_viewport_size(self, width: int, height: int) -> None:
        self._page.set_viewport_size({"width": width, "height": height})

    async def get_viewport_size(self) -> ViewPortSize:
        width, height, scroll_height = await self.run_js(
            "[window.innerWidth, window.innerHeight, document.documentElement.scrollHeight]"
        )

        return {"width": width, "height": height, "content_height": scroll_height}


class PlaywrightAsyncAdapter(BrowserAdapter):
    STRIP_RETURN = "return window."

    def __init__(self, page: PageAsync):
        self._page = page

    async def run_js(self, js: str) -> Any:
        if js.startswith(self.STRIP_RETURN):
            js = js[len(self.STRIP_RETURN) :]

        return await self._page.evaluate(js)

    async def take_screenshot(self) -> bytes:
        return await self._page.screenshot(type="png")

    async def set_viewport_size(self, width: int, height: int) -> None:
        await self._page.set_viewport_size({"width": width, "height": height})

    async def get_viewport_size(self) -> ViewPortSize:
        width, height, scroll_height = await self.run_js(
            "[window.innerWidth, window.innerHeight, document.documentElement.scrollHeight]"
        )

        return {"width": width, "height": height, "content_height": scroll_height}



import asyncio
from typing import Any

from selenium.webdriver.remote.webdriver import WebDriver

from tarsier.adapter._base import BrowserAdapter
from tarsier.adapter.types import ViewPortSize


class SeleniumAdapter(BrowserAdapter):
    def __init__(self, driver: WebDriver):
        self.driver = driver

    async def run_js(self, js: str) -> Any:
        return self.driver.execute_script(js)

    async def take_screenshot(self) -> bytes:
        return self.driver.get_screenshot_as_png()

    async def set_viewport_size(self, width: int, height: int) -> None:
        self.driver.set_window_size(width, height)

    async def get_viewport_size(self) -> ViewPortSize:
        script = "return [window.innerWidth, window.innerHeight, document.documentElement.scrollHeight];"
        width, height, content_height = self.driver.execute_script(script)
        return {"width": width, "height": height, "content_height": content_height}



from os import PathLike
from typing import Any


def load_js(path: PathLike[Any]) -> str:
    try:
        with open(path, "r") as f:
            return f.read()
    except FileNotFoundError as e:
        raise ValueError(
            "Could not find tag_utils.js. Please ensure that you complied typescript using `npm run build`"
        ) from e



from abc import ABC, abstractmethod
from typing import Any

from tarsier.adapter.types import ViewPortSize


class BrowserAdapter(ABC):
    @abstractmethod
    async def run_js(self, js: str) -> Any:
        pass

    @abstractmethod
    async def take_screenshot(self) -> bytes:
        pass

    @abstractmethod
    async def set_viewport_size(self, width: int, height: int) -> None:
        pass

    @abstractmethod
    async def get_viewport_size(self) -> ViewPortSize:
        pass



from asyncio import Protocol
from typing import Dict, Tuple

from tarsier.adapter import AnyDriver, BrowserAdapter, adapter_factory


class ITarsier(Protocol):
    async def page_to_image(self, driver: AnyDriver) -> Tuple[bytes, Dict[int, str]]:
        raise NotImplementedError()

    async def page_to_text(self, driver: AnyDriver) -> Tuple[str, Dict[int, str]]:
        raise NotImplementedError()
